# (c) 2016 Anaconda, Inc. / https://anaconda.com
# All Rights Reserved
#
# constructor is distributed under the terms of the BSD 3-clause license.
# Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.

import os
import platform
import shutil
import sys
import time
from os.path import isdir, join
from os.path import split as path_split
from pathlib import Path
from textwrap import dedent
from typing import List, Mapping, Union

from . import __version__ as CONSTRUCTOR_VERSION
from .conda_interface import (
    CONDA_INTERFACE_VERSION,
    Dist,
    MatchSpec,
    PrefixData,
    all_channel_urls,
    default_prefix,
)
from .conda_interface import distro as conda_distro
from .conda_interface import get_repodata, write_repodata
from .utils import ensure_transmuted_ext, filename_dist, get_final_channels, get_final_url

try:
    import json
except ImportError:
    import ruamel_json as json

files = '.constructor-build.info', 'urls', 'urls.txt', 'env.txt'


def write_index_cache(info, dst_dir, used_packages):
    cache_dir = join(dst_dir, 'cache')

    if not isdir(cache_dir):
        os.makedirs(cache_dir)

    _platforms = info['_platform'], 'noarch'
    _remap_configs = list(info.get("channels_remap", []))
    _env_channels = []
    for env_info in info.get("extra_envs", {}).values():
        _remap_configs += env_info.get("channels_remap", [])
        _env_channels += env_info.get("channels", [])

    _remaps = {url['src'].rstrip('/'): url['dest'].rstrip('/')
               for url in _remap_configs}
    _channels = [
        url.rstrip("/")
        for url in list(_remaps)
        + info.get("channels", [])
        + info.get("conda_default_channels", [])
        + _env_channels
    ]
    _urls = all_channel_urls(_channels, subdirs=_platforms)
    repodatas = {url: get_repodata(url) for url in _urls if url is not None}

    all_urls = info["_urls"].copy()
    for env_info in info.get("_extra_envs_info", {}).values():
        all_urls += env_info["_urls"]

    for url, _ in all_urls:
        src, subdir, fn = url.rsplit('/', 2)
        dst = _remaps.get(src)
        if dst is not None:
            src = '%s/%s' % (src, subdir)
            dst = '%s/%s' % (dst, subdir)
            if dst not in repodatas:
                repodatas[dst] = {
                    '_url': dst,
                    'info': {'subdir': subdir},
                    'packages': {},
                    'packages.conda': {},
                    'removed': []
                }
            loc = 'packages.conda' if fn.endswith('.conda') else 'packages'
            repodatas[dst][loc][fn] = repodatas[src][loc][fn]
    for src in _remaps:
        for subdir in _platforms:
            del repodatas['%s/%s' % (src, subdir)]

    for url, repodata in repodatas.items():
        if repodata is not None:
            write_repodata(cache_dir, url, repodata, used_packages, info)

    for cache_file in os.listdir(cache_dir):
        if not cache_file.endswith(".json"):
            os.unlink(join(cache_dir, cache_file))


def system_info():
    out = {'constructor': CONSTRUCTOR_VERSION,
           'conda': CONDA_INTERFACE_VERSION,
           'platform': sys.platform,
           'python': sys.version,
           'python_version': tuple(sys.version_info),
           'machine': platform.machine(),
           'platform_full': platform.version()}
    if sys.platform == 'darwin':
        out['extra'] = platform.mac_ver()
    elif sys.platform.startswith('linux'):
        if conda_distro is not None:
            out['extra'] = conda_distro.linux_distribution(full_distribution_name=False)
        elif hasattr(platform, 'dist'):
            out['extra'] = platform.dist()
    elif sys.platform.startswith('win'):
        out['extra'] = platform.win32_ver()
        prefix = default_prefix
        prefix_records = list(PrefixData(prefix).iter_records())
        nsis_prefix_rec = next(
            (rec for rec in prefix_records if rec.name == 'nsis'), None)
        if nsis_prefix_rec:
            out['nsis'] = nsis_prefix_rec.version
    return out


def write_files(info, dst_dir):
    with open(join(dst_dir, '.constructor-build.info'), 'w') as fo:
        json.dump(system_info(), fo)

    all_urls = info["_urls"].copy()
    for env_info in info.get("_extra_envs_info", {}).values():
        all_urls += env_info["_urls"]

    final_urls_md5s = tuple((get_final_url(info, url), md5) for url, md5 in info["_urls"])
    all_final_urls_md5s = tuple((get_final_url(info, url), md5) for url, md5 in all_urls)

    with open(join(dst_dir, 'urls'), 'w') as fo:
        for url, md5 in all_final_urls_md5s:
            maybe_different_url = ensure_transmuted_ext(info, url)
            if maybe_different_url != url:  # transmuted, no md5
                fo.write(f"{maybe_different_url}\n")
            else:
                fo.write(f"{url}#{md5}\n")

    with open(join(dst_dir, 'urls.txt'), 'w') as fo:
        for url, _ in all_final_urls_md5s:
            fo.write('%s\n' % url)

    all_dists = info["_dists"].copy()
    for env_info in info.get("_extra_envs_info", {}).values():
        all_dists += env_info["_dists"]
    all_dists = list({dist: None for dist in all_dists})  # de-duplicate

    write_index_cache(info, dst_dir, all_dists)

    # base environment conda-meta
    write_conda_meta(info, dst_dir, final_urls_md5s)

    write_repodata_record(info, dst_dir)

    # base environment file used with conda install --file
    # (list of specs/dists to install)
    write_env_txt(info, dst_dir, final_urls_md5s)

    for fn in files:
        os.chmod(join(dst_dir, fn), 0o664)

    for env_name, env_info in info.get("_extra_envs_info", {}).items():
        env_config = info["extra_envs"][env_name]
        env_dst_dir = os.path.join(dst_dir, "envs", env_name)
        # environment conda-meta
        env_urls_md5 = tuple((get_final_url(info, url), md5) for url, md5 in env_info["_urls"])
        user_requested_specs = env_config.get('user_requested_specs', env_config.get('specs', ()))
        write_conda_meta(info, env_dst_dir, env_urls_md5, user_requested_specs)
        # environment installation list
        write_env_txt(info, env_dst_dir, env_urls_md5)
        # channels
        write_channels_txt(info, env_dst_dir, env_config)


def write_conda_meta(info, dst_dir, final_urls_md5s, user_requested_specs=None):
    if user_requested_specs is None:
        user_requested_specs = info.get('user_requested_specs', info.get('specs', ()))

    cmd = path_split(sys.argv[0])[-1]
    if len(sys.argv) > 1:
        cmd = "%s %s" % (cmd, " ".join(sys.argv[1:]))

    builder = [
        "==> %s <==" % time.strftime('%Y-%m-%d %H:%M:%S'),
        "# cmd: %s" % cmd,
    ]
    dists = tuple(Dist(url) for url, _ in final_urls_md5s)

    builder.extend("+%s" % dist.full_name for dist in dists)
    if user_requested_specs:
        update_specs = [str(MatchSpec(s)) for s in user_requested_specs]
        builder.append("# update specs: %s" % update_specs)
    builder.append("\n")

    if not isdir(join(dst_dir, 'conda-meta')):
        os.makedirs(join(dst_dir, 'conda-meta'))
    with open(join(dst_dir, 'conda-meta', 'history'), 'w') as fh:
        fh.write("\n".join(builder))


def write_repodata_record(info, dst_dir):
    all_dists = info["_dists"].copy()
    for env_data in info.get("_extra_envs_info", {}).values():
        all_dists += env_data["_dists"]
    for dist in all_dists:
        if filename_dist(dist).endswith(".conda"):
            _dist = filename_dist(dist)[:-6]
        elif filename_dist(dist).endswith(".tar.bz2"):
            _dist = filename_dist(dist)[:-8]
        record_file = join(_dist, 'info', 'repodata_record.json')
        record_file_src = join(info['_download_dir'], record_file)

        with open(record_file_src) as rf:
            rr_json = json.load(rf)

        rr_json['url'] = get_final_url(info, rr_json['url'])
        rr_json['channel'] = get_final_url(info, rr_json['channel'])

        if not isdir(join(dst_dir, _dist, 'info')):
            os.makedirs(join(dst_dir, _dist, 'info'))

        record_file_dest = join(dst_dir, record_file)

        with open(record_file_dest, 'w') as rf:
            json.dump(rr_json, rf, indent=2, sort_keys=True)


def write_env_txt(info, dst_dir, urls):
    """
    urls is an iterable of tuples with url and md5 values
    """
    header = dedent(
        f"""
        # This file may be used to create an environment using:
        # $ conda create --name <env> --file <this file>
        # platform: {info['_platform']}
        @EXPLICIT
        """
    ).lstrip()
    with open(join(dst_dir, "env.txt"), "w") as envf:
        envf.write(header)
        for url, md5 in urls:
            maybe_different_url = ensure_transmuted_ext(info, url)
            if maybe_different_url != url:  # transmuted, no md5
                envf.write(f"{maybe_different_url}\n")
            else:
                envf.write(f"{url}#{md5}\n")


def write_channels_txt(info, dst_dir, env_config):
    env_config = env_config.copy()
    if "channels" not in env_config:
        env_config["channels"] = info.get("channels", ())
    if "channels_remap" not in env_config:
        env_config["channels_remap"] = info.get("channels_remap", ())

    with open(join(dst_dir, "channels.txt"), "w") as f:
        f.write(",".join(get_final_channels(env_config)))


def copy_extra_files(
    extra_files: List[Union[os.PathLike, Mapping]], workdir: os.PathLike
) -> List[os.PathLike]:
    """Copy list of extra files to a working directory

    Args:
        extra_files (List[Union[os.PathLike, Mapping]]): A path or a mapping
        workdir (os.PathLike): Path to where extra files will be copied to.

    Raises:
        FileNotFoundError: Raises when the file isn't found.

    Returns:
        List[os.PathLike]: List of normalized paths of copied locations.
    """
    if not extra_files:
        return []
    copied = []
    for path in extra_files:
        if isinstance(path, str):
            copied.append(shutil.copy(path, workdir))
        elif isinstance(path, dict):
            assert len(path) == 1
            origin, destination = next(iter(path.items()))
            orig_path = Path(origin)
            if not orig_path.exists():
                raise FileNotFoundError(f"File {origin} does not exist.")
            dest_path = Path(workdir) / destination
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            copied.append(shutil.copy(orig_path, dest_path))
    return copied



# (c) 2016 Anaconda, Inc. / https://anaconda.com
# All Rights Reserved
#
# constructor is distributed under the terms of the BSD 3-clause license.
# Consult LICENSE.txt or http://opensource.org/licenses/BSD-3-Clause.


import argparse
import logging
import os
import sys
from os.path import abspath, expanduser, isdir, join
from textwrap import dedent, indent

from . import __version__
from .build_outputs import process_build_outputs
from .conda_interface import SUPPORTED_PLATFORMS, cc_platform
from .construct import generate_key_info_list, ns_platform
from .construct import parse as construct_parse
from .construct import verify as construct_verify
from .fcp import main as fcp_main
from .utils import normalize_path, yield_lines

DEFAULT_CACHE_DIR = os.getenv('CONSTRUCTOR_CACHE', '~/.conda/constructor')

logger = logging.getLogger(__name__)


def get_installer_type(info):
    osname, unused_arch = info['_platform'].split('-')

    os_allowed = {'linux': ('sh',), 'osx': ('sh', 'pkg'), 'win': ('exe',)}
    all_allowed = set(sum(os_allowed.values(), ('all',)))

    itype = info.get('installer_type')
    if not itype:
        return os_allowed[osname][:1]
    elif itype == 'all':
        return os_allowed[osname]
    elif itype not in all_allowed:
        all_allowed = ', '.join(sorted(all_allowed))
        sys.exit("Error: invalid installer type '%s'; allowed: %s" % (itype, all_allowed))
    elif itype not in os_allowed[osname]:
        os_allowed = ', '.join(sorted(os_allowed[osname]))
        sys.exit("Error: invalid installer type '%s' for %s; allowed: %s" %
                 (itype, osname, os_allowed))
    else:
        return itype,


def get_output_filename(info):
    try:
        return info['installer_filename']
    except KeyError:
        pass

    osname, arch = info['_platform'].split('-')
    os_map = {'linux': 'Linux', 'osx': 'MacOSX', 'win': 'Windows'}
    arch_name_map = {'64': 'x86_64', '32': 'x86'}
    ext = info['installer_type']
    return '%s-%s-%s.%s' % ('%(name)s-%(version)s' % info,
                            os_map.get(osname, osname),
                            arch_name_map.get(arch, arch),
                            ext)


def main_build(dir_path, output_dir='.', platform=cc_platform,
               verbose=True, cache_dir=DEFAULT_CACHE_DIR,
               dry_run=False, conda_exe="conda.exe"):
    logger.info('platform: %s', platform)
    if not os.path.isfile(conda_exe):
        sys.exit("Error: Conda executable '%s' does not exist!" % conda_exe)
    cache_dir = abspath(expanduser(cache_dir))
    try:
        osname, unused_arch = platform.split('-')
    except ValueError:
        sys.exit("Error: invalid platform string '%s'" % platform)

    construct_path = join(dir_path, 'construct.yaml')
    info = construct_parse(construct_path, platform)
    construct_verify(info)
    info['CONSTRUCTOR_VERSION'] = __version__
    info['_input_dir'] = dir_path
    info['_output_dir'] = output_dir
    info['_platform'] = platform
    info['_download_dir'] = join(cache_dir, platform)
    info['_conda_exe'] = abspath(conda_exe)
    itypes = get_installer_type(info)

    if platform != cc_platform and 'pkg' in itypes and not cc_platform.startswith('osx-'):
        sys.exit("Error: cannot construct a macOS 'pkg' installer on '%s'" % cc_platform)
    if osname == "win" and "micromamba" in os.path.basename(info['_conda_exe']):
        # TODO: Remove when shortcut creation is implemented on micromamba
        sys.exit("Error: micromamba is not supported on Windows installers.")

    logger.debug('conda packages download: %s', info['_download_dir'])

    for key in ('welcome_image_text', 'header_image_text'):
        if key not in info:
            info[key] = info['name']

    for key in ('license_file', 'welcome_image', 'header_image', 'icon_image',
                'pre_install', 'post_install', 'pre_uninstall', 'environment_file',
                'nsis_template', 'welcome_file', 'readme_file', 'conclusion_file',
                'signing_certificate'):
        if info.get(key):  # only join if there's a truthy value set
            info[key] = abspath(join(dir_path, info[key]))

    for key in 'specs', 'packages':
        if key not in info:
            continue
        if isinstance(info[key], str):
            info[key] = list(yield_lines(join(dir_path, info[key])))

    # normalize paths to be copied; if they are relative, they must be to
    # construct.yaml's parent (dir_path)
    extras_types = ['extra_files', 'temp_extra_files']
    for extra_type in extras_types:
        extras = info.get(extra_type, ())
        new_extras = []
        for path in extras:
            if isinstance(path, str):
                new_extras.append(abspath(join(dir_path, path)))
            elif isinstance(path, dict):
                assert len(path) == 1
                orig, dest = next(iter(path.items()))
                orig = abspath(join(dir_path, orig))
                new_extras.append({orig: dest})
        info[extra_type] = new_extras

    for key in 'channels', 'specs', 'exclude', 'packages', 'menu_packages':
        if key in info:
            # ensure strings in those lists are stripped
            info[key] = [line.strip() for line in info[key]]
            # ensure there are no empty strings
            if any((not s) for s in info[key]):
                sys.exit("Error: found empty element in '%s:'" % key)

    for env_name, env_config in info.get("extra_envs", {}).items():
        if env_name in ("base", "root"):
            raise ValueError(f"Environment name '{env_name}' cannot be used")
        for config_key, value in env_config.copy().items():
            if isinstance(value, (list, tuple)):
                env_config[config_key] = [val.strip() for val in value]
            if config_key == "environment_file":
                env_config[config_key] = abspath(join(dir_path, value))

    info['installer_type'] = itypes[0]
    fcp_main(info, verbose=verbose, dry_run=dry_run, conda_exe=conda_exe)
    if dry_run:
        logger.info("Dry run, no installers or build outputs created.")
        return

    # info has keys
    # 'name', 'version', 'channels', 'exclude',
    # '_platform', '_download_dir', '_outpath'
    # 'specs': ['python 3.5*', 'conda', 'nomkl', 'numpy', 'scipy', 'pandas',
    #           'notebook', 'matplotlib', 'lighttpd']
    # 'license_file': '/Users/kfranz/continuum/constructor/examples/miniconda/EULA.txt'
    # '_dists': List[Dist]
    # '_urls': List[Tuple[url, md5]]

    for itype in itypes:
        if itype == 'sh':
            from .shar import create as shar_create
            create = shar_create
        elif itype == 'pkg':
            from .osxpkg import create as osxpkg_create
            create = osxpkg_create
        elif itype == 'exe':
            from .winexe import create as winexe_create
            create = winexe_create
        info['installer_type'] = itype
        info['_outpath'] = abspath(join(output_dir, get_output_filename(info)))
        create(info, verbose=verbose)
        logger.info("Successfully created '%(_outpath)s'.", info)

    process_build_outputs(info)


class _HelpConstructAction(argparse.Action):
    def __init__(
        self,
        option_strings,
        dest=argparse.SUPPRESS,
        default=argparse.SUPPRESS,
        help="describe available configuration options for construct.yaml files and exit",
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help,
        )

    def __call__(self, parser, namespace, values, option_string=None):

        parser._print_message(self._build_message(), sys.stdout)
        parser.exit()

    def _build_message(self):
        msg = dedent(
            """
            The 'construct.yaml' specification
            ==================================

            constructor version {version}

            The `construct.yaml` file is the primary mechanism for controlling
            the output of the Constructor package. The file contains a list of
            key/value pairs in the standard YAML format.

            Available keys
            --------------

            {available_keys}

            Available selectors
            -------------------

            Constructor can use the same Selector enhancement of the YAML format
            used in conda-build ('# [selector]'). Available keywords are:

            {available_selectors}
            """
        )
        available_keys_list = []
        for key, required, key_types, help_msg, plural in generate_key_info_list():
            available_keys_list.append(
                "\n".join(
                    [
                        key,
                        "·" * len(key),
                        indent(
                            f"Required: {required}, type{plural}: {key_types}", "    "
                        ),
                        indent(help_msg.strip(), "    "),
                        "",
                    ]
                )
            )
        available_selectors_list = [
            f"- {sel}" for sel in sorted(ns_platform(sys.platform).keys())
        ]
        return msg.format(
            version=__version__,
            available_keys="\n".join(available_keys_list),
            available_selectors="\n".join(available_selectors_list),
        )


def main():
    logging.basicConfig(level=logging.INFO)
    p = argparse.ArgumentParser(
        description="build an installer from <DIRECTORY>/construct.yaml")

    p.add_argument("--help-construct", action=_HelpConstructAction)

    p.add_argument('--debug',
                   action="store_true")

    p.add_argument('--output-dir',
                   action="store",
                   default=os.getcwd(),
                   help='path to directory in which output installer is written '
                   "to, defaults to CWD ('{}')".format(os.getcwd()),
                   metavar='PATH')

    p.add_argument('--cache-dir',
                   action="store",
                   default=DEFAULT_CACHE_DIR,
                   help='cache directory, used for downloading conda packages, '
                   'may be changed by CONSTRUCTOR_CACHE, '
                   "defaults to '{}'".format(DEFAULT_CACHE_DIR),
                   metavar='PATH')

    p.add_argument('--clean',
                   action="store_true",
                   help='clean out the cache directory and exit')

    p.add_argument('--platform',
                   action="store",
                   default=cc_platform,
                   help="the platform for which installer is for, "
                   f"defaults to '{cc_platform}'. Options, e.g.: {SUPPORTED_PLATFORMS}")

    p.add_argument('--dry-run',
                   help="solve package specs but do not create installer",
                   default=False,
                   action="store_true")

    p.add_argument('-v', '--verbose',
                   action="store_true")

    p.add_argument('-V', '--version',
                   help="display the version being used and exit",
                   action="version",
                   version=f'%(prog)s {__version__}')

    p.add_argument('--conda-exe',
                   help="path to conda executable (conda-standalone, micromamba)",
                   action="store",
                   metavar="CONDA_EXE")

    p.add_argument('dir_path',
                   help="directory containing construct.yaml",
                   action="store",
                   nargs="?",
                   default=os.getcwd(),
                   metavar='DIRECTORY')

    args = p.parse_args()
    logger.info("Got the following cli arguments: '%s'", args)

    if args.verbose or args.debug:
        logging.getLogger("constructor").setLevel(logging.DEBUG)

    if args.clean:
        import shutil
        cache_dir = abspath(expanduser(args.cache_dir))
        logger.info("cleaning cache: '%s'", cache_dir)
        if isdir(cache_dir):
            shutil.rmtree(cache_dir)
        return

    dir_path = args.dir_path
    if not isdir(dir_path):
        p.error("no such directory: %s" % dir_path)

    conda_exe = args.conda_exe
    conda_exe_default_path = os.path.join(sys.prefix, "standalone_conda", "conda.exe")
    conda_exe_default_path = normalize_path(conda_exe_default_path)
    if conda_exe:
        conda_exe = normalize_path(os.path.abspath(conda_exe))
    elif args.platform != cc_platform:
        p.error("setting --conda-exe is required for building a non-native installer")
    else:
        conda_exe = conda_exe_default_path
    if not os.path.isfile(conda_exe):
        if conda_exe != conda_exe_default_path:
            p.error("file not found: %s" % args.conda_exe)
        p.error("""
no standalone conda executable was found. The
easiest way to obtain one is to install the 'conda-standalone' package.
Alternatively, you can download an executable manually and supply its
path with the --conda-exe argument. Self-contained executables can be
downloaded from https://repo.anaconda.com/pkgs/misc/conda-execs/""".lstrip())

    out_dir = normalize_path(args.output_dir)
    main_build(dir_path, output_dir=out_dir, platform=args.platform,
               verbose=args.verbose, cache_dir=args.cache_dir,
               dry_run=args.dry_run, conda_exe=conda_exe)


if __name__ == '__main__':
    main()
